@@ -16,6 +16,8 @@ use HTTP::Status;
 use MIME::Base64;
 use REST::Client;
 use URI::Escape;
+use LWP::UserAgent;
+use XML::Simple;
 use Kernel::Config;
 
 use Kernel::System::VariableCheck qw(:all);
@@ -76,6 +78,9 @@ In case of an error, the resulting http error code and message are remembered fo
 sub ProviderProcessRequest {
     my ( $Self, %Param ) = @_;
 
+    my $JSONObject  = $Kernel::OM->Get('Kernel::System::JSON');
+    my $ParamObject = $Kernel::OM->Get('Kernel::System::Web::Request');
+
     # Check transport config.
     if ( !IsHashRefWithData( $Self->{TransportConfig} ) ) {
         return $Self->_Error(
@@ -154,6 +159,9 @@ sub ProviderProcessRequest {
         }
     }
 
+    my $ParserBackend;
+    my %ParserBackendParameter;
+
     my $RequestMethod = $ENV{'REQUEST_METHOD'} || 'GET';
     ROUTE:
     for my $CurrentOperation ( sort keys %{ $Config->{RouteOperationMapping} } ) {
@@ -198,6 +206,12 @@ sub ProviderProcessRequest {
             $URIData{$URIKey} = $URIValue;
         }
 
+        $ParserBackend = $RouteMapping{ParserBackend} || 'JSON';
+
+        if ( IsHashRefWithData( $RouteMapping{ParserBackendParameter} ) ) {
+            %ParserBackendParameter = %{ $RouteMapping{ParserBackendParameter} };
+        }
+
         $Operation = $CurrentOperation;
 
         # Leave with the first matching regexp.
@@ -209,6 +223,21 @@ sub ProviderProcessRequest {
         %URIData = ( %QueryParams, %URIData, );
     }
 
+    my %DataHeaderOverwrite = (
+        HTTP_X_OTRS_HEADER_USERLOGIN         => 'UserLogin',
+        HTTP_X_OTRS_HEADER_CUSTOMERUSERLOGIN => 'CustomerUserLogin',
+        HTTP_X_OTRS_HEADER_SESSIONID         => 'SessionID',
+        HTTP_X_OTRS_HEADER_PASSWORD          => 'Password',
+        HTTP_X_OTRS_HEADER_IMPERSONATEASUSER => 'ImpersonateAsUser',
+    );
+
+    HEADER:
+    for my $Header ( sort keys %DataHeaderOverwrite ) {
+        next HEADER if !IsStringWithData( $ENV{$Header} );
+
+        $URIData{ $DataHeaderOverwrite{$Header} } = $ENV{$Header};
+    }
+
     if ( !$Operation ) {
         return $Self->_Error(
             Summary   => "HTTP::REST Error while determine Operation for request URI '$RequestURI'.",
@@ -240,9 +269,36 @@ sub ProviderProcessRequest {
         );
     }
 
-    # Read request.
-    my $Content;
-    read STDIN, $Content, $Length;
+    # try to read request from special CGI parameters
+    # those are filled:
+    # 'If POSTed data is not of type application/x-www-form-urlencoded or multipart/form-data'
+    # https://metacpan.org/pod/CGI#Handling-non-urlencoded-arguments
+    my $Content = $ParamObject->GetParam( Param => 'POSTDATA' )
+        // $ParamObject->GetParam( Param => 'PUTDATA' )
+        // $ParamObject->GetParam( Param => 'PATCHDATA' );
+
+    # if no content is given yet, maybe there were some
+    # 'application/x-www-form-urlencoded or multipart/form-data'
+    # submitted...
+    if ( !IsStringWithData($Content) ) {
+
+        # get possible POST parameter name from config
+        # otherwise fall back on 'file'
+        my %UploadInfo = $Kernel::OM->Get('Kernel::System::Web::Request')->GetUploadAll(
+            Param => $Config->{PostParamName} || 'file',
+        );
+
+        # if the parameter indeed has some content with it
+        # pass it to $Content so we can work with it later
+        if ( IsStringWithData( $UploadInfo{Content} ) ) {
+            $Content = $UploadInfo{Content};
+        }
+    }
+
+    # fallback to otrs standard
+    if ( !$Content ) {
+        read STDIN, $Content, $Length;
+    }
 
     # If there is no STDIN data it might be caused by fastcgi already having read the request.
     # In this case we need to get the data from CGI.
@@ -261,9 +317,13 @@ sub ProviderProcessRequest {
         );
     }
 
+    # remove utf-8 BOMs:
+    # http://www.ueber.net/who/mjl/projects/bomstrip/
+    $Content =~ s{\xef\xbb\xbf}{}g;
+
     # Convert char-set if necessary.
     my $ContentCharset;
-    if ( $ENV{'CONTENT_TYPE'} =~ m{ \A .* charset= ["']? ( [^"']+ ) ["']? \z }xmsi ) {
+    if ( $ENV{'CONTENT_TYPE'} =~ m{ \A .* charset\s*=\s* ["']? ( [^"']+ ) ["']? \z }xmsi ) {
         $ContentCharset = $1;
     }
     if ( $ContentCharset && $ContentCharset !~ m{ \A utf [-]? 8 \z }xmsi ) {
@@ -282,9 +342,19 @@ sub ProviderProcessRequest {
         Data    => $Content,
     );
 
-    my $ContentDecoded = $Kernel::OM->Get('Kernel::System::JSON')->Decode(
-        Data => $Content,
-    );
+    my $ContentDecoded;
+    if ( $ParserBackend eq 'JSON' ) {
+        $ContentDecoded = $JSONObject->Decode(
+            %ParserBackendParameter,
+            Data => $Content,
+        );
+    }
+    else {
+        return $Self->_Error(
+            Summary   => "Invalid parser backend '$ParserBackend' configured for Operation '$Operation'.",
+            HTTPError => 500,
+        );
+    }
 
     if ( !$ContentDecoded ) {
         return $Self->_Error(
@@ -339,7 +409,7 @@ The HTTP code is set accordingly
 - C<500> for content syntax errors
 
     my $Result = $TransportObject->ProviderGenerateResponse(
-        Success => 1
+        Success => 1,
         Data    => { # data payload for response, optional
             ...
         },
@@ -363,16 +433,9 @@ sub ProviderGenerateResponse {
         );
     }
 
-    # Check data param.
-    if ( defined $Param{Data} && ref $Param{Data} ne 'HASH' ) {
-        return $Self->_Output(
-            HTTPCode => 500,
-            Content  => 'Invalid data',
-        );
-    }
-
     # Check success param.
-    my $HTTPCode = 200;
+    my $HTTPCode = $Param{HTTPCode} // 200;
+
     if ( !$Param{Success} ) {
 
         # Create Fault structure.
@@ -382,13 +445,23 @@ sub ProviderGenerateResponse {
             faultstring => $FaultString,
         };
 
-        # Override HTTPCode to 500.
-        $HTTPCode = 500;
+        # overide HTTPCode to 500 if no custom HTTP code was provided
+        $HTTPCode = 500 if !$Param{HTTPCode};
     }
 
-    # Orepare data.
+    my %AdditionalResponseHeaders;
+    if ( exists $Param{Data}->{AdditionalResponseHeaders} ) {
+        if ( IsHashRefWithData( $Param{Data}->{AdditionalResponseHeaders} ) ) {
+            %AdditionalResponseHeaders = %{ $Param{Data}->{AdditionalResponseHeaders} };
+        }
+
+        delete $Param{Data}->{AdditionalResponseHeaders};
+    }
+
+    # Prepare data.
     my $JSONString = $Kernel::OM->Get('Kernel::System::JSON')->Encode(
-        Data => $Param{Data},
+        Data     => $Param{Data},
+        SortKeys => 1,
     );
 
     if ( !$JSONString ) {
@@ -400,8 +473,9 @@ sub ProviderGenerateResponse {
 
     # No error - return output.
     return $Self->_Output(
-        HTTPCode => $HTTPCode,
-        Content  => $JSONString,
+        HTTPCode                  => $HTTPCode,
+        Content                   => $JSONString,
+        AdditionalResponseHeaders => \%AdditionalResponseHeaders,
     );
 }
 
@@ -430,6 +504,10 @@ receive the response and return its data.
 sub RequesterPerformRequest {
     my ( $Self, %Param ) = @_;
 
+    my $EncodeObject      = $Kernel::OM->Get('Kernel::System::Encode');
+    my $JSONObject        = $Kernel::OM->Get('Kernel::System::JSON');
+    my $OAuth2TokenObject = $Kernel::OM->Get('Kernel::System::OAuth2Token');
+
     # Check transport config.
     if ( !IsHashRefWithData( $Self->{TransportConfig} ) ) {
         return {
@@ -475,12 +553,74 @@ sub RequesterPerformRequest {
     my $Headers = { 'Content-Type' => 'application/json; charset=UTF-8' };
 
     # set up a REST session
-    my $RestClient = REST::Client->new(
-        {
+    my $RestClient = $Kernel::OM->{Objects}->{ $Config->{Host} };
+
+    if (
+        !$Config->{ReuseClient}
+        || !$RestClient
+        )
+    {
+        # fill default parameter
+        my %RESTClientParameter = (
             host    => $Config->{Host},
             timeout => $Config->{Timeout},
+        );
+
+        # if configured we have to skip hostname verification
+        # on ssl connections for e.g. self signed certs
+        if ( $Config->{SSLNoHostnameVerification} ) {
+
+            # therefore we have to use our own LWP::UserAgent object
+            # with disabled 'verify_hostname' SSL option
+            $RESTClientParameter{useragent} = LWP::UserAgent->new(
+                keep_alive => $Config->{KeepAlive},
+                ssl_opts   => {
+                    verify_hostname => 0,
+                    SSL_verify_mode => 0,
+                },
+            );
         }
-    );
+
+        # create a REST::Client object we can perform our session with
+        $RestClient = REST::Client->new( \%RESTClientParameter );
+
+        if ( !$RestClient ) {
+            my $ErrorMessage = "Error while creating REST client from 'REST::Client'.";
+
+            # log to debugger
+            $Self->{DebuggerObject}->Error(
+                Summary => $ErrorMessage,
+            );
+            return {
+                Success      => 0,
+                ErrorMessage => $ErrorMessage,
+            };
+        }
+
+        # add X509 options if configured
+        if ( IsHashRefWithData( $Config->{X509} ) ) {
+
+            # use X509 options
+            if (
+                IsStringWithData( $Config->{X509}->{UseX509} )
+                && $Config->{X509}->{UseX509} eq 'Yes'
+                )
+            {
+                #X509 client authentication
+                $RestClient->setCert( $Config->{X509}->{X509CertFile} );
+                $RestClient->setKey( $Config->{X509}->{X509KeyFile} );
+
+                #add a CA to verify server certificates
+                if ( IsStringWithData( $Config->{X509}->{X509CAFile} ) ) {
+                    $RestClient->setCa( $Config->{X509}->{X509CAFile} );
+                }
+            }
+        }
+
+        if ( $Config->{ReuseClient} ) {
+            $Kernel::OM->{Objects}->{ $Config->{Host} } = $RestClient;
+        }
+    }
 
     if ( !$RestClient ) {
 
@@ -562,12 +702,78 @@ sub RequesterPerformRequest {
                 )
             {
                 $Headers->{'Proxy-Authorization'} = 'Basic ' . encode_base64(
-                    $Config->{Proxy}->{ProxyUser} . ':' . $Config->{Proxy}->{ProxyPassword}
+                    $Config->{Proxy}->{ProxyUser} . ':' . $Config->{Proxy}->{ProxyPassword},
+                    '',
                 );
             }
         }
     }
 
+    #
+    # JWT support
+    #
+    my $JWTObject             = $Kernel::OM->Get('Kernel::System::JSONWebToken');
+    my $X509CertificateObject = $Kernel::OM->Get('Kernel::System::X509Certificate');
+
+    my $JWTObjectIsSupported             = $JWTObject->IsSupported();
+    my $X509CertificateObjectIsSupported = $X509CertificateObject->IsSupported();
+
+    my $JWT;
+
+    if (
+        $JWTObjectIsSupported
+        && IsHashRefWithData( $Config->{Authentication} )
+        && IsStringWithData( $Config->{Authentication}->{AuthType} )
+        && $Config->{Authentication}->{AuthType} eq 'JWT'
+        )
+    {
+        my %JWTPlaceholderData;
+
+        # Fetch data from X.509 certificate, if configured and supported
+        # to be able to insert it into configured placeholders of payload and additional header data
+        # of the JWT.
+        if (
+            $X509CertificateObjectIsSupported
+            && $Config->{Authentication}->{JWTAuthCertificateFilePath}
+            )
+        {
+            my $X509Certificate = $X509CertificateObject->Parse(
+                FilePath => $Config->{Authentication}->{JWTAuthCertificateFilePath},
+            ) // {};
+
+            for my $X509CertificateKey ( sort keys %{$X509Certificate} ) {
+                my $X509CertificateValue = $X509Certificate->{$X509CertificateKey} // '';
+                $JWTPlaceholderData{ 'OTRS_JWT_Cert' . $X509CertificateKey } = $X509CertificateValue;
+            }
+        }
+
+        # Calculate expiration date and insert it into placeholders of payload and additional header data
+        # of the JWT.
+        my $TTL            = int( $Config->{Authentication}->{JWTAuthTTL} );
+        my $DateTimeObject = $Kernel::OM->Create(
+            'Kernel::System::DateTime',
+            ObjectParams => {
+                TimeZone => 'UTC',
+            }
+        );
+        $DateTimeObject->Add( Seconds => $TTL );
+        $JWTPlaceholderData{OTRS_JWT_ExpirationDateTimestamp} = $DateTimeObject->ToEpoch();
+        $JWTPlaceholderData{OTRS_JWT_ExpirationDateString}    = $DateTimeObject->Format(
+            Format => '%Y-%m-%dT%H:%M:%S %{time_zone_long_name}',
+        );
+
+        my $Payload              = $Config->{Authentication}->{JWTAuthPayload}              // {};
+        my $AdditionalHeaderData = $Config->{Authentication}->{JWTAuthAdditionalHeaderData} // {};
+        $JWT = $JWTObject->Encode(
+            Payload              => $Payload,
+            Algorithm            => $Config->{Authentication}->{JWTAuthAlgorithm},
+            KeyFilePath          => $Config->{Authentication}->{JWTAuthKeyFilePath},
+            KeyPassword          => $Config->{Authentication}->{JWTAuthKeyFilePassword},
+            AdditionalHeaderData => $AdditionalHeaderData,
+            PlaceholderData      => \%JWTPlaceholderData,
+        );
+    }
+
     # Add authentication options if configured (hard wired to basic authentication at the moment).
     if (
         IsHashRefWithData( $Config->{Authentication} )
@@ -578,10 +784,23 @@ sub RequesterPerformRequest {
         )
     {
         $Headers->{Authorization} = 'Basic ' . encode_base64(
-            $Config->{Authentication}->{BasicAuthUser} . ':' . $Config->{Authentication}->{BasicAuthPassword}
+            $Config->{Authentication}->{BasicAuthUser} . ':' . $Config->{Authentication}->{BasicAuthPassword},
+            '',
         );
     }
 
+    # JWT header
+    elsif (
+        $JWTObjectIsSupported
+        && IsHashRefWithData( $Config->{Authentication} )
+        && IsStringWithData( $Config->{Authentication}->{AuthType} )
+        && $Config->{Authentication}->{AuthType} eq 'JWT'
+        && IsStringWithData($JWT)
+        )
+    {
+        $Headers->{Authorization} = "Bearer $JWT";
+    }
+
     my $RestCommand = $Config->{DefaultCommand};
     if ( IsStringWithData( $Config->{InvokerControllerMapping}->{ $Param{Operation} }->{Command} ) )
     {
@@ -624,6 +843,35 @@ sub RequesterPerformRequest {
         };
     }
 
+    if ( grep { $_ eq $RestCommand } qw(POST PUT PATCH) ) {
+        if ( IsStringWithData( $Config->{ContentType} ) ) {
+
+            # Add defined content-type to HTTP header
+            if ( $Config->{ContentType} eq 'JSON' ) {
+                $Headers->{'Content-Type'} = 'application/json';
+            }
+            elsif ( $Config->{ContentType} eq 'FORM' ) {
+                $Headers->{'Content-Type'} = 'application/x-www-form-urlencoded';
+            }
+            elsif ( $Config->{ContentType} eq 'XML' ) {
+                $Headers->{'Content-Type'} = 'application/xml';
+            }
+        }
+    }
+
+    my $RequestHeaders = delete $Param{Data}->{__RequestHeaders};
+    if ( IsHashRefWithData($RequestHeaders) ) {
+        $Headers = $RequestHeaders;
+    }
+
+    my $RequestHeadersAppend = delete $Param{Data}->{__RequestHeadersAppend};
+    if ( IsHashRefWithData($RequestHeadersAppend) ) {
+        %{$Headers} = (
+            %{$Headers},
+            %{$RequestHeadersAppend},
+        );
+    }
+
     my @RequestParam;
     my $Controller = $Config->{InvokerControllerMapping}->{ $Param{Operation} }->{Controller};
 
@@ -637,16 +885,34 @@ sub RequesterPerformRequest {
 
     my @ParamsToDelete;
 
+    my %FlattenedParamData;
+    $Self->_FlattenDataStructure(
+        Data      => $Param{Data},
+        Flattened => \%FlattenedParamData,
+    );
+
     # Replace any URI params with their actual value.
-    #    for example: from /Ticket/:TicketID/:Other
-    #    to /Ticket/1/2 (considering that $Param{Data} contains TicketID = 1 and Other = 2).
-    for my $ParamName ( sort keys %{ $Param{Data} } ) {
-        if ( $Controller =~ m{:$ParamName(?=/|\?|$)}msx ) {
-            my $ParamValue = $Param{Data}->{$ParamName};
-            $ParamValue = URI::Escape::uri_escape_utf8($ParamValue);
-            $Controller =~ s{:$ParamName(?=/|\?|$)}{$ParamValue}msxg;
-            push @ParamsToDelete, $ParamName;
-        }
+    #    for example: from /Ticket/:TicketData.TicketID/:Other
+    #    to /Ticket/1/2 (considering that $Param{Data} contains {TicketData}->{TicketID} = 1 and {Other} = 2).
+    FLATTENEDPARAMDATA:
+    for my $FlattenedParamName ( sort keys %FlattenedParamData ) {
+        next FLATTENEDPARAMDATA if $Controller !~ m{:$FlattenedParamName(?=/|\?|$)}ms;
+
+        my $ParamValue = $FlattenedParamData{$FlattenedParamName};
+        next FLATTENEDPARAMDATA if !defined $ParamValue;
+
+        $EncodeObject->EncodeInput( \$ParamValue );
+
+        $ParamValue = URI::Escape::uri_escape_utf8($ParamValue);
+        $Controller =~ s{:$FlattenedParamName(?=/|\?|$)}{$ParamValue}msg;
+
+        # Only delete "top level" hash keys.
+        # Keep data of nested structures as is because it would
+        # look like there's something missing otherwise (especially if
+        # array elements would be removed).
+        next FLATTENEDPARAMDATA if index( $FlattenedParamName, '.' ) != -1;
+
+        push @ParamsToDelete, $FlattenedParamName;
     }
 
     $Self->{DebuggerObject}->Debug(
@@ -661,14 +927,26 @@ sub RequesterPerformRequest {
         #    to ?UserLogin=user&Password=secret
         #    (considering that $Param{Data} contains UserLogin = 'user' and Password = 'secret').
         my $ReplaceFlag;
-        for my $ParamName ( sort keys %{ $Param{Data} } ) {
-            if ( $QueryParamsStr =~ m{:$ParamName(?=&|$)}msx ) {
-                my $ParamValue = $Param{Data}->{$ParamName};
-                $ParamValue = URI::Escape::uri_escape_utf8($ParamValue);
-                $QueryParamsStr =~ s{:$ParamName(?=&|$)}{$ParamValue}msxg;
-                push @ParamsToDelete, $ParamName;
-                $ReplaceFlag = 1;
-            }
+        FLATTENEDPARAMDATA:
+        for my $FlattenedParamName ( sort keys %FlattenedParamData ) {
+            next FLATTENEDPARAMDATA if $QueryParamsStr !~ m{:$FlattenedParamName(?=&|$)}ms;
+
+            my $ParamValue = $FlattenedParamData{$FlattenedParamName};
+            next FLATTENEDPARAMDATA if !defined $ParamValue;
+
+            $EncodeObject->EncodeInput( \$ParamValue );
+
+            $ParamValue = URI::Escape::uri_escape_utf8($ParamValue);
+            $QueryParamsStr =~ s{:$FlattenedParamName(?=&|$)}{$ParamValue}msxg;
+            $ReplaceFlag = 1;
+
+            # Only delete "top level" hash keys.
+            # Keep data of nested structures as is because it would
+            # look like there's something missing otherwise (especially if
+            # array elements would be removed).
+            next FLATTENEDPARAMDATA if index( $FlattenedParamName, '.' ) != -1;
+
+            push @ParamsToDelete, $FlattenedParamName;
         }
 
         # Append query params in the URI.
@@ -682,15 +960,11 @@ sub RequesterPerformRequest {
         }
     }
 
-    # Remove already used params.
+    # Remove already used "top-level" params.
     for my $ParamName (@ParamsToDelete) {
         delete $Param{Data}->{$ParamName};
     }
 
-    # Get JSON and Encode object.
-    my $JSONObject   = $Kernel::OM->Get('Kernel::System::JSON');
-    my $EncodeObject = $Kernel::OM->Get('Kernel::System::Encode');
-
     my $Body;
     if ( IsHashRefWithData( $Param{Data} ) ) {
 
@@ -706,9 +980,49 @@ sub RequesterPerformRequest {
                 Data    => $Param{Data},
             );
 
-            $Param{Data} = $JSONObject->Encode(
-                Data => $Param{Data},
-            );
+            if ( $Headers->{'Content-Type'} eq 'application/x-www-form-urlencoded' ) {
+
+                # Add parameters as form data to body.
+                $Param{Data} = $RestClient->buildQuery(
+                    %{ $Param{Data} }
+                );
+
+                # Remove any leading questing mark
+                if ( IsStringWithData( $Param{Data} ) ) {
+                    $Param{Data} =~ s{\A\?}{}sm;
+                }
+            }
+            elsif ( $Headers->{'Content-Type'} eq 'application/xml' ) {
+                my $XMLSimple = XML::Simple->new();
+                my $XMLOut;
+                eval {
+                    $XMLOut = $XMLSimple->XMLout(
+                        $Param{Data},
+                        AttrIndent => 1,
+                        ContentKey => '-content',
+                        NoAttr     => 0,
+                        KeyAttr    => [],
+                        RootName   => undef,
+                    );
+                };
+
+                $XMLOut //= '';
+
+                if ( !length $XMLOut ) {
+                    $Self->{DebuggerObject}->Error(
+                        Summary => 'Could not generate XML content for request.',
+                    );
+                }
+
+                $Param{Data} = '<?xml version="1.0" encoding="UTF-8"?>' . "\n" . $XMLOut;
+            }
+            else {
+
+                # Default to JSON (OTRS default behavior)
+                $Param{Data} = $JSONObject->Encode(
+                    Data => $Param{Data},
+                );
+            }
 
             # Make sure data is correctly encoded.
             $EncodeObject->EncodeOutput( \$Param{Data} );
@@ -748,11 +1062,92 @@ sub RequesterPerformRequest {
         push @RequestParam, $Body;
     }
 
+    if ( IsHashRefWithData( $Config->{AdditionalHeaders} ) ) {
+        my $AdditionalHeaders = $Config->{AdditionalHeaders};
+
+        # Insert data into placeholders.
+        my %PlaceholderData;
+        if (
+            IsHashRefWithData( $Config->{Authentication} )
+            && IsStringWithData( $Config->{Authentication}->{AuthType} )
+            )
+        {
+
+            # JWT
+            if (
+                $Config->{Authentication}->{AuthType} eq 'JWT'
+                && $JWTObjectIsSupported
+                && IsStringWithData($JWT)
+                )
+            {
+                $PlaceholderData{OTRS_JWT} = $JWT;
+            }
+
+            # OAuth2 token
+            if (
+                $Config->{Authentication}->{AuthType} eq 'OAuth2Token'
+                && $Config->{Authentication}->{OAuth2TokenConfigID}
+                )
+            {
+                my $OAuth2Token = $OAuth2TokenObject->GetToken(
+                    TokenConfigID => $Config->{Authentication}->{OAuth2TokenConfigID},
+                    UserID        => 1,
+                );
+                if ( !IsStringWithData($OAuth2Token) ) {
+                    my $ErrorMessage
+                        = "OAuth2 token for config with ID $Config->{Authentication}->{OAuth2TokenConfigID} could not be retrieved.";
+                    $Self->{DebuggerObject}->Error(
+                        Summary => $ErrorMessage,
+                    );
+                    return {
+                        Success      => 0,
+                        ErrorMessage => $ErrorMessage,
+                    };
+                }
+
+                $PlaceholderData{OTRS_OAUTH2_TOKEN} = $OAuth2Token;
+            }
+        }
+
+        HEADERFIELDNAME:
+        for my $HeaderFieldname ( sort keys %{$AdditionalHeaders} ) {
+            my $HeaderValue = $AdditionalHeaders->{$HeaderFieldname};
+            next HEADERFIELDNAME if !defined $HeaderValue;
+
+            for my $Placeholder ( sort keys %PlaceholderData ) {
+                my $PlaceholderValue = $PlaceholderData{$Placeholder} // '';
+                $HeaderValue =~ s{<$Placeholder>}{$PlaceholderValue}g;
+            }
+
+            # Remove unknown placeholders.
+            $HeaderValue =~ s{<OTRS_.*?>}{}g;
+
+            $AdditionalHeaders->{$HeaderFieldname} = $HeaderValue;
+        }
+
+        %{$Headers} = (
+            %{$Headers},
+            %{ $Config->{AdditionalHeaders} },
+        );
+    }
+
+    if ( IsHashRefWithData($Headers) ) {
+        for my $HeaderName ( sort keys %{$Headers} ) {
+            my $HeaderValue = $Headers->{$HeaderName};
+            $RestClient->addHeader( $HeaderName, $HeaderValue );
+        }
+    }
+
     # Add headers to request
     push @RequestParam, $Headers;
 
     $RestClient->$RestCommand(@RequestParam);
 
+    $Self->{DebuggerObject}->Debug(
+        Summary => "Request params from outgoing data",
+        Data    => \@RequestParam,
+    );
+
     my $ResponseCode = $RestClient->responseCode();
     my $ResponseError;
     my $ErrorMessage = "Error while performing REST '$RestCommand' request to Controller '$Controller' on Host '"
@@ -767,6 +1162,12 @@ sub RequesterPerformRequest {
     }
 
     my $ResponseContent = $RestClient->responseContent();
+
+    $Self->{DebuggerObject}->Debug(
+        Summary => "JSON data received from remote system",
+        Data    => $ResponseContent,
+    );
+
     if ( $ResponseCode ne '204' && !IsStringWithData($ResponseContent) ) {
         $ResponseError .= ' No content provided.';
     }
@@ -792,7 +1193,7 @@ sub RequesterPerformRequest {
     }
 
     # Send processed data to debugger.
-    my $SizeExeeded = 0;
+    my $SizeExceeded = 0;
     {
         my $MaxSize = $Kernel::OM->Get('Kernel::Config')->Get('GenericInterface::Operation::ResponseLoggingMaxSize')
             || 200;
@@ -808,7 +1209,7 @@ sub RequesterPerformRequest {
             );
         }
         else {
-            $SizeExeeded = 1;
+            $SizeExceeded = 1;
             $Self->{DebuggerObject}->Debug(
                 Summary => "JSON data received from remote system was too large for logging",
                 Data =>
@@ -827,6 +1228,19 @@ sub RequesterPerformRequest {
 
     if ( $ResponseCode ne '204' ) {
 
+        my $ResponseContentType = $RestClient->responseHeader('Content-Type') // '';
+        if ( $ResponseContentType =~ m{text/(plain|html|xml)} ) {
+            $Result = {
+                Body => $ResponseContent,
+            };
+
+            return {
+                Success      => 1,
+                Data         => $Result,
+                SizeExceeded => $SizeExceeded,
+            };
+        }
+
         $Result = $JSONObject->Decode(
             Data => $ResponseContent,
         );
@@ -847,9 +1261,9 @@ sub RequesterPerformRequest {
 
     # All OK - return result.
     return {
-        Success     => 1,
-        Data        => $Result || undef,
-        SizeExeeded => $SizeExeeded,
+        Success      => 1,
+        Data         => $Result || undef,
+        SizeExceeded => $SizeExceeded,
     };
 }
 
@@ -933,12 +1347,17 @@ sub _Output {
 
     # prepare additional headers
     my $AdditionalHeaderStrg = '';
+
+    my %AdditionalHeaders;
     if ( IsHashRefWithData( $Self->{TransportConfig}->{Config}->{AdditionalHeaders} ) ) {
-        my %AdditionalHeaders = %{ $Self->{TransportConfig}->{Config}->{AdditionalHeaders} };
-        for my $AdditionalHeader ( sort keys %AdditionalHeaders ) {
-            $AdditionalHeaderStrg
-                .= $AdditionalHeader . ': ' . ( $AdditionalHeaders{$AdditionalHeader} || '' ) . "\r\n";
-        }
+        %AdditionalHeaders = %{ $Self->{TransportConfig}->{Config}->{AdditionalHeaders} };
+    }
+    if ( IsHashRefWithData( $Param{AdditionalResponseHeaders} ) ) {
+        %AdditionalHeaders = ( %AdditionalHeaders, %{ $Param{AdditionalResponseHeaders} } );
+    }
+    for my $AdditionalHeader ( sort keys %AdditionalHeaders ) {
+        $AdditionalHeaderStrg
+            .= $AdditionalHeader . ': ' . ( $AdditionalHeaders{$AdditionalHeader} // '' ) . "\r\n";
     }
 
     # In the constructor of this module STDIN and STDOUT are set to binmode without any additional
@@ -1017,6 +1436,112 @@ sub _Error {
     };
 }
 
+=head2 _FlattenDataStructure()
+
+    Returns a flattened hash/array of a given hash/array.
+
+    $TransportObject->_FlattenDataStructure(
+        Data => \%OldHash,
+        Flattened => \%NewHash,
+    );
+
+    my %OldHash = (
+        Config => {
+            A => 1,
+            B => 2,
+            C => 3,
+            D => [
+                2, 5, 6,
+            ],
+        },
+        Config2 => 1
+    );
+
+    my %NewHash = (
+        Config.A   => 1,
+        Config.B   => 1,
+        Config.C   => 1,
+        Config.D.0 => 2,
+        Config.D.1 => 5,
+        Config.D.2 => 6,
+        Config2    => 1,
+    );
+
+=cut
+
+sub _FlattenDataStructure {
+    my ( $Self, %Param ) = @_;
+
+    for my $Needed (qw(Data Flattened)) {
+        if ( !$Param{$Needed} ) {
+            print "Got no $Needed!\n";
+            return;
+        }
+    }
+
+    my @Containers;
+    my $DataType;
+
+    if ( IsHashRefWithData( $Param{Data} ) ) {
+        @Containers = sort keys %{ $Param{Data} };
+        $DataType   = 'Hash';
+    }
+    else {
+        @Containers = @{ $Param{Data} };
+        $DataType   = 'Array';
+    }
+
+    my $Prefix = $Param{Prefix} // '';
+
+    my $ArrayCount = 0;
+
+    CONTAINER:
+    for my $Container (@Containers) {
+        next CONTAINER if !$Container;
+
+        if ( $DataType eq 'Hash' ) {
+            if (
+                IsHashRefWithData( $Param{Data}->{$Container} )
+                || IsArrayRefWithData( $Param{Data}->{$Container} )
+                )
+            {
+                $Self->_FlattenDataStructure(
+                    Data      => $Param{Data}->{$Container},
+                    Flattened => $Param{Flattened},
+                    Prefix    => $Prefix . $Container . '.',
+                );
+            }
+            else {
+                $Prefix                      = $Prefix . $Container;
+                $Param{Flattened}->{$Prefix} = $Param{Data}->{$Container};
+                $Prefix                      = $Param{Prefix} || '';
+            }
+        }
+        else {
+            if (
+                IsHashRefWithData($Container)
+                || IsArrayRefWithData($Container)
+                )
+            {
+                $Self->_FlattenDataStructure(
+                    Data      => $Container,
+                    Flattened => $Param{Flattened},
+                    Prefix    => $Prefix . $Container . '.',
+                );
+            }
+            else {
+                $Prefix                      = $Prefix . $ArrayCount;
+                $Param{Flattened}->{$Prefix} = $Container;
+                $Prefix                      = $Param{Prefix} || '';
+            }
+
+            $ArrayCount++;
+        }
+    }
+
+    return 1;
+}
+
 1;
 
 =end Internal:
